static const char* const rcsid = "$Id$";
/*                               -*- Mode: C -*- 
 * oprs-mrPost.c -- 
 * 
 * Copyright (C) 1994-2005 LAAS/CNRS
 * 
 * Author          : Francois Felix Ingrand
 * Created On      : Tue Apr 26 11:37:29 1994
 * Last Modified By: Francois Felix Ingrand
 * Last Modified On: Tue Jul 25 12:11:21 2000
 * Update Count    : 114
 * Status          : OK
 * 
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>

#include <hashPack.h>

#include "macro.h"

#include "opaque-pub.h"
#include "macro-pub.h"
#undef FREE

#include "oprs-type-pub.h"
#include "pu-mk-term_f.h"
#include "pu-parse-tl_f.h"


#include "portLib.h"
#include "csLib.h"
#include "posterLib.h"
#include "h2errorLib.h"

#include "oprs-post_f.h"

#include "oprs-type_f-pub.h"
#include "lisp-list_f-pub.h"
#include "user-decl-oprs-post.h"
#include "user-decl-oprs-post_f.h"

#include "user-decl-oprs-com.h"

#define POSTER_TAKE_TIME_OUT 50

#define COPY_LIST(list) copy_list((list),NOT_NODES)
#define FREE_LIST(list) free_list((list),NOT_NODES)

typedef struct {		
     char *name;
     POSTER_ID poster_id;
     void *addr;
     PBoolean created_by_me;
} Poster;

typedef enum {DECODE, ENCODE} Poster_Function_Type;

typedef struct {
     char *name;
     Encode_Func_Proto encode;
     Decode_Func_Proto decode;
     char *poster_name;
     Poster *poster;
     int offset;
} Poster_Function;


#define POSTER_HASH_SIZE 32		/* has to be a power of two */
static int poster_hashtable_size = POSTER_HASH_SIZE;
static int poster_hashtable_mod = POSTER_HASH_SIZE - 1;
HASH poster_hash;

#define POSTER_FUNCTION_HASH_SIZE 128		/* has to be a power of two */
static int poster_function_hashtable_size = POSTER_FUNCTION_HASH_SIZE;
static int poster_function_hashtable_mod = POSTER_FUNCTION_HASH_SIZE -1;
HASH poster_function_hash;

static int poster_hash_a_string(PString name )	
{
     return (hash_a_string(name) & poster_hashtable_mod);
}

static PBoolean poster_match(PString p1, Poster *poster)
{
     return ( p1 == poster->name) ;
}

static int poster_function_hash_a_string(PString name )	
{
     return (hash_a_string(name) & poster_function_hashtable_mod);
}

static PBoolean poster_function_match(PString f1, Poster_Function *poster_function)
{
     return ( f1 == poster_function->name) ;
}

static int delete_poster(void *ignore, Poster *poster)
{
     if (poster->created_by_me)
	  if (posterDelete(poster->poster_id) == ERROR) {
	       h2perror("oprs-com:delete_poster");
	  }
     return 1;
}

void end_oprs_post()
{
     for_all_hashtable(poster_hash, NULL, delete_poster);
}

void init_oprs_post()
{
     poster_hash = make_hashtable(poster_hashtable_size, poster_hash_a_string, poster_match);
     poster_function_hash = make_hashtable(poster_function_hashtable_size,
					   poster_function_hash_a_string,
					   poster_function_match);
}

static char *get_string_from_id_bar(char * id)
{
     char *res;

     if (id[0] == '|') {
	  int len = strlen(id);
	  if ((char *)strchr(id + 1,'|') != (id + len - 1))	/* if the second | is not the last character... */
	  	  fprintf(stderr, "Bad id in get_string_from_id_bar.\n"); 
	  res = (char *)malloc(len-1);
	  memcpy(res, id + 1, len-2);
	  res[len-2] = '\0';
     } else
	  NEW_STRING(id,res);

     return res;
}

static int oprs_poster_create(char *poster_name, int i)
{
     Poster *poster = MAKE_OBJECT(Poster);
     char *ext_name;

     poster->name = poster_name;

     ext_name = get_string_from_id_bar(poster->name);

     if (posterCreate(ext_name, i, &poster->poster_id) == ERROR) {
	  h2perror("oprs-com:oprs_poster_create:mrPosterCreate");
	  return -1;
     }

     if ((poster->addr = posterAddr(poster->poster_id)) == NULL) {
	  fprintf(stderr, "oprs-com:oprs_poster_create:NULL addr.\n"); 
	  return -1;
     }

     poster->created_by_me = TRUE;

     if ((get_from_hashtable(poster_hash,poster->name)) != NULL)
	  fprintf(stderr, "oprs-com:oprs_poster_create:redeclaring the poster: %s.\n", poster_name); 
     
     add_to_hashtable(poster_hash, poster, poster->name); /* redeclaring it anyway... */

     free(ext_name);

     return 0;
}

static Poster *oprs_find_poster(char *poster_name)
{
     Poster *poster;
     char *ext_name;

     if ((poster = (Poster *)get_from_hashtable(poster_hash,poster_name)) != NULL)
	  return poster;

     poster = MAKE_OBJECT(Poster);

     poster->name = poster_name;	/* Declare the sub system name symbol */

     ext_name = get_string_from_id_bar(poster->name);

     if ((posterFind(ext_name, &poster->poster_id )) != OK) {
	  h2perror("oprs-com:oprs_poster_find_poster:posterFind");
	  return NULL;
     }

     if ((poster->addr = posterAddr(poster->poster_id)) == NULL) {
	  h2perror("oprs-com:oprs_find_poster:posterAddr:NULL addr"); 
	  return NULL;
     }
     poster->created_by_me = FALSE;

     add_to_hashtable(poster_hash, poster, poster_name);

     free(ext_name);

     return poster;
}

void declare_poster_function(char *poster_name, char *name, Decode_Func_Proto decode,
						Encode_Func_Proto encode, int offset)
{
     Poster_Function *poster_function = MAKE_OBJECT(Poster_Function);
     
     poster_function->name = declare_atom(name);	/* Declare the poster_function as a symbol */
     poster_function->encode = encode;
     poster_function->decode = decode;
     poster_function->offset = offset;

     poster_function->poster_name = declare_atom(poster_name);	/* Declare the poster_function as a symbol */
     poster_function->poster = NULL;
	  
     if ((get_from_hashtable(poster_function_hash, poster_function->name)) != NULL)
	  fprintf(stderr, "oprs-com: The poster function \"%s\" has already been declared.\n", poster_function->name); 

     add_to_hashtable(poster_function_hash, poster_function, poster_function->name); 

     return;
}   


Term *create_poster(TermList terms)
{
     Term *t1, *t2, *res;

     res = MAKE_OBJECT(Term);
     res->type = ATOM;

     t1 = (Term *)get_list_pos(terms,1);
     t2 = (Term *)get_list_pos(terms,2);
     if (t2->type != INTEGER || t1->type != ATOM) {
	  fprintf(stderr,"Expecting an ATOM and an INTEGER in create_poster.\n");
	  res->u.id = nil_sym;
     } else {
	  if (oprs_poster_create(t1->u.id, t2->u.intval) < 0)
	       res->u.id = nil_sym;
	  else
	       res->u.id = lisp_t_sym;
     }
     return res;
}

Term *write_poster(TermList terms)
/* This is the action which sends the request. */
{
     Poster_Function *poster_function;
     int encode_res;
     PBoolean param_res;
     Term * res;
     char *function_name;
     Expression *term;
     STATUS err;

     if (! (param_res = PUGetOprsParameters(terms, 2, ATOM, &function_name, EXPRESSION, &term))) {
	  fprintf(stderr, "oprs-com:write_poster: unable to parse request.\n"); 
	  res = build_nil();
     } else if ((poster_function = (Poster_Function *)get_from_hashtable(poster_function_hash,
									 function_name)) == NULL) {
	  fprintf(stderr, "oprs-com:write_poster: could not poster_function %s.\n", function_name); 
	  res = build_nil();
     } else if (!(poster_function->poster || (poster_function->poster = oprs_find_poster(poster_function->poster_name)))) {
	  fprintf(stderr, "oprs-com:write_poster: could not find poster_id for function %s with poster %s.\n",
		  poster_function->name, poster_function->poster_name); 
	  res = build_nil();
     } else if ((err = posterTake(poster_function->poster->poster_id, POSTER_WRITE)) == ERROR) {
	  h2perror("oprs-com:write_poster:posterTake");
	  res = build_nil();
     } 
     else if (err != OK) {
	  res = build_nil();
     } else {
	  encode_res = (*(poster_function->encode))(NULL,
						    term,
						    (char *)(poster_function->poster->addr)+poster_function->offset,
							    1);
	  if ((err = posterGive(poster_function->poster->poster_id) ) == ERROR) {
	       h2perror("oprs-com:write_poster:posterGive");
	       res = build_nil();
	  } else {

	       if (! encode_res) {
		    fprintf(stderr, "oprs-com:poster_write:problem while encoding poster data\n\
\tfor function %s with poster %s.\n", poster_function->name, poster_function->poster_name); 
		    res = build_nil();
	       } else 
		    res = build_t();
	  }
     }
     return res;
}
     
Term *read_poster(TermList terms)
/* This is the action which sends the request. */
{
     Poster_Function *poster_function;
     Term * res;
     PBoolean param_res;
     char *function_name;
     STATUS err;

     if (! (param_res = PUGetOprsParameters(terms, 1, ATOM, &function_name))) {
	  fprintf(stderr, "oprs-com:read_poster: unable to parse request with data.\n"); 
	  return  build_nil();
     } else  if ((poster_function = (Poster_Function *)get_from_hashtable(poster_function_hash,
									  function_name)) == NULL) {
	  fprintf(stderr, "oprs-com:read_poster: could not poster_function %s.\n", function_name); 
	  res = build_nil();
     } else if (!(poster_function->poster || (poster_function->poster = oprs_find_poster(poster_function->poster_name)))) {
	  fprintf(stderr, "oprs-com:read_poster: could not find poster_id for function %s with poster %s.\n",
		  poster_function->name, poster_function->poster_name); 
	  res = build_nil();
     } else  if ((err = posterTake(poster_function->poster->poster_id, POSTER_READ) ) == ERROR) {
	  h2perror("oprs-com:read_poster:posterTake");
	  res = build_nil();
     } else if (err != OK) {
	  res = build_nil();
     } else {
	  res = (*(poster_function->decode))(NULL,(char *)(poster_function->poster->addr)+poster_function->offset,1);

	  if ((err = posterGive(poster_function->poster->poster_id) ) == ERROR) {
	       h2perror("oprs-com:read_poster:posterGive");
	       res = build_nil();
	  }
     }
     return res;
}

     
PBoolean poster_exist(TermList terms)
/* This is the action which sends the request. */
{
     char *poster_name;

     if (! PUGetOprsParameters(terms, 1, ATOM, &poster_name)) {
	       fprintf(stderr, "oprs-com:poster_exist: unable to parse request.\n"); 
	       return FALSE;
     }
     if (oprs_find_poster(poster_name)) {
	  return TRUE;
     } else {
	  return FALSE;
     }
}


int collect_poster(OPRS_LIST res, Poster *poster)
{
     add_to_head(res, build_id(poster->name));
     return 1;
}

Term *ef_list_posters(TermList terms)
{
     OPRS_LIST res = make_list();

     sh_for_all_hashtable(poster_hash, res, collect_poster);
     
     return build_l_list(transform_c_list_to_l_list(res));
}


int collect_poster_function(OPRS_LIST res, Poster_Function *poster_function)
{
     add_to_head(res, build_id(poster_function->name));
     return 1;
}

Term *ef_list_poster_functions(TermList terms)
{
     OPRS_LIST res = make_list();

     sh_for_all_hashtable(poster_function_hash, res, collect_poster_function);
     
     return build_l_list(transform_c_list_to_l_list(res));
}

